import magpack
from magtomo.Experiment import Experiment, get_electric_field
import numpy as np
import logging
import functools
import multiprocessing
from scipy.linalg import block_diag
from scipy.ndimage import affine_transform


class Reconstruction(Experiment):
    """Performs a reconstruction from dichroic projections, inherits the properties that are specified by an
    :class:`magtomo.Experiment.Experiment`."""

    def __init__(self, magnetization, rotations, projections, pol=1j, iterations=10, mask=None, learning_parameter=10):
        """
        Parameters
        ----------
        magnetization : np.ndarray
            The initial guess for the magnetization.
        rotations : np.ndarray
            Stack of rotation matrices describing the sample orientations that were measured; shaped (n, 3, 3).
        projections : np.ndarray
            Stack of projections obtained at the corresponding `rot` value.
        pol : array_like (optional)
            Single polarization used for all projections, stack of polarizations matching the number of projections or
            stack of polarization pairs. Linear polarization represented using values from 0 to 180 (in degrees).
            Circular left and right polarization represented using complex ±1j.
        iterations : int (optional)
            Number of reconstruction iterations to perform.
        mask : np.ndarray (optional)
            Binary mask for the magnetization.
        learning_parameter : float (optional)
            The learning parameter for gradient descent.
        """
        super().__init__(magnetization, rotations, pol)
        self._error_metric = None
        self.iterations = iterations
        self.learning_parameter = learning_parameter
        self.projections = projections
        self.mask = mask
        self._error_norm = self.calc_error_norm()

    @property
    def projections(self):
        """The measured projections."""
        return self._projections

    @projections.setter
    def projections(self, value):
        if value.ndim != 3:
            raise ValueError("Projections must be 3-dimensional.")
        self._projections = value
        self.calc_error_norm()

    @property
    def learning_parameter(self):
        """The learning parameter for gradient descent."""
        return self.learning_parameter

    @learning_parameter.setter
    def learning_parameter(self, value):
        self._learning_parameter = value

    @property
    def iterations(self):
        """The number of reconstruction iterations."""
        return self._iterations

    @iterations.setter
    def iterations(self, value):
        if value < 1:
            raise ValueError("Iterations must be greater than 0.")
        self._iterations = value
        if self._error_metric is None or self._error_metric.shape != self._iterations:
            self._error_metric = np.empty(self._iterations)

    @property
    def error_metric(self):
        """The error metric value at each iteration."""
        return self._error_metric

    @property
    def mask(self):
        """Binary mask for the magnetization."""
        return self._mask

    @mask.setter
    def mask(self, value: np.ndarray):
        if value is None:
            self._mask = np.ones_like(self._magnetization)
        elif value.shape == self.magnetization.shape[-3:]:
            self._mask = value
        else:
            raise ValueError("Mask must be same shape as magnetization.")

    def calc_error_norm(self):
        """Calculates a normalization factor (number of pixels in the projections) for scaling the error metric.

        Returns
        -------
        int
            The normalization factor (number of pixels in the projections).
        """
        return np.prod(self._projections.shape)

    def configuration_error(self, reference, axial=False):
        r"""Computes the error between the current reconstruction and a provided reference.

        Parameters
        ----------
        reference : np.ndarray
            Reference structure as a numpy array that will be compared with the current reconstruction.
        axial : bool (optional)
            Set to True if reconstructing orientation fields (e.g. linear dichroic projections).

        Returns
        -------
        float
            The configuration error.

        Notes
        -----
        The configuration error, :math:`\varepsilon_c` is evaluated for directional (vector) and axial (orientation)
        fields using:

        .. math::

            \varepsilon_c = \begin{cases} \vec{m}_r - \vec{m} & \text{directional} \\
            \min\left(\left\lvert \vec{m}_r - \vec{m} \right\rvert,
            \left\lvert \vec{m}_r + \vec{m} \right\rvert\right) & \text{axial}
            \end{cases}

        where :math:`\vec{m}_r` is the reconstruction and :math:`\vec{m}` is the reference.
        """
        if reference.shape != self._magnetization.shape:
            raise ValueError("Reference shape must match magnetization shape.")

        error = np.sum((self._magnetization - reference) ** 2, axis=0)
        if axial:
            error_branch = np.sum((self._magnetization + reference) ** 2, axis=0)
            error = np.where(error_branch < error, error_branch, error)
        return error

    def theta_error(self, reference: np.ndarray, axial: bool = False):
        r"""Computes the angular difference between the current reconstruction and a provided reference.

        Parameters
        ----------
        reference : np.ndarray
            Reference structure as a numpy array that will be compared with the current reconstruction.
        axial : bool (optional)
            Set to True if reconstructing orientation fields (e.g. linear dichroic projections).

        Returns
        -------
        float
            The angular error in radians.

        Notes
        -----
        The angular error, :math:`\varepsilon_\theta` is evaluated for directional (vector) and axial (orientation)
        fields using:

        .. math::

            \varepsilon_c =
            \begin{cases} \arccos\left( \frac{\vec{m_r} \cdot \vec{m}}{|\vec{m_r}||\vec{m}|} \right)
            & \text{directional} \\
            \arccos\left( \frac{\left\lvert \vec{m_r} \cdot \vec{m} \right\rvert}{|\vec{m_r}| |\vec{m}|} \right)
             & \text{axial}
            \end{cases}

        where :math:`\vec{m}_r` is the reconstruction and :math:`\vec{m}` is the reference.
        """
        return magpack.vectorop.angular_difference(self._magnetization, reference, axial=axial)

    def reconstruct(self):
        """Performs the reconstructions and updates the magnetization."""
        self._prepare()
        learning_param = self._learning_parameter

        for ii in range(self._iterations):
            logging.info(f"Currently on iteration {ii + 1}")

            # compute differences
            self.calculate_sinogram()
            difference = self._projections - self._sinogram
            current_error = 2 * np.sum(difference ** 2) / self._error_norm
            self._error_metric[ii] = current_error
            logging.info(f"Error metric: {current_error}")

            # evaluate gradients
            if np.all(self._pol.imag == 0):
                grad_method = tensor_gradient
                if not self._is_tensor:
                    self.orientation_to_tensor()
            else:
                grad_method = gradient

            partial_gradient = functools.partial(grad_method, self._magnetization, order=self._order)
            with multiprocessing.Pool() as p:
                grad = p.starmap(partial_gradient, zip(self._rotations, difference, self._pol))

            # apply gradient
            total_gradient = np.sum(grad, axis=0) / self._rotations.shape[0]
            dx = total_gradient / np.linalg.norm(total_gradient)
            self._magnetization += learning_param * dx * self._mask

    def _prepare(self):
        """Perform checks before initiating reconstruction"""
        if isinstance(self._pol, (int, float, complex)):
            self.pol = np.repeat(self._pol, self._rotations.shape[0])
        elif isinstance(self._pol, np.ndarray) and self._pol.shape[0] != self._rotations.shape[0]:
            raise ValueError("Number of polarizations must match number of rotations.")
        self.magnetization = self._magnetization * self._mask


def _compute_difference_field(field, difference, rot, order):
    """Compute the rotated difference field."""
    diff_field = difference[..., np.newaxis].repeat(field.shape[-1], axis=-1) / field.shape[-1]
    return rotate(diff_field, rot, order)


def tensor_gradient(tf, rot, difference, pol, order):
    """Calculates the tensor gradient for a specific orientation.

    Parameters
    ----------
    tf : np.ndarray
        The current guess for the tensor field.
    rot : np.ndarray
        The rotation matrix describing the sample rotation
    difference : np.ndarray
        The difference between the projection from `vf` and the input.
    pol : np.ndarray
        The polarization at which the projection was measured.
    order : int
        The interpolation order for performing rotations.

    Returns
    -------
    np.ndarray
        The gradient for this combination of inputs.
    """
    difference_field = _compute_difference_field(tf, difference, rot, order)

    electric_field = (np.cos(np.deg2rad(pol)), np.sin(np.deg2rad(pol)), 0)
    grad = np.einsum('i,im,kn,k', electric_field, rot, rot, electric_field)
    output_grad = np.einsum('ij,abc->ijabc', grad, difference_field)
    return output_grad


def gradient(vf, rot, difference, pol, order):
    """Calculates the gradient for a specific orientation.

    Parameters
    ----------
    vf : np.ndarray
        The current guess for the vector or orientation field.
    rot : np.ndarray
        The rotation matrix describing the sample rotation
    difference : np.ndarray
        The difference between the projection from `vf` and the input.
    pol : np.ndarray
        The polarization at which the projection was measured.
    order : int
        The interpolation order for performing rotations.

    Returns
    -------
    np.ndarray
        The gradient for this combination of inputs.
    """
    if isinstance(pol, np.ndarray) and len(pol) == 2:
        return gradient(vf, rot, difference, pol[0], order) - gradient(vf, rot, difference, pol[1], order)
    elif isinstance(pol, np.ndarray):
        raise ValueError("Polarizations for dichroic input must be in pairs.")

    difference_field = _compute_difference_field(vf, difference, rot, order)

    electric_field = get_electric_field(pol)
    if electric_field[2] != 0:
        grad_x, grad_y, grad_z = rot[2] * electric_field[2]
    else:
        grad_x, grad_y, grad_z = np.einsum('ij,lk,jabc,i,l->kabc', rot, rot, vf, electric_field,
                                           electric_field) * 4
    grad_x = np.multiply(grad_x, difference_field)
    grad_y = np.multiply(grad_y, difference_field)
    grad_z = np.multiply(grad_z, difference_field)

    output_grad = np.stack([grad_x, grad_y, grad_z])
    return output_grad


def rotate(config, rot_matrix, order=1):
    """Applies an affine rotation described by the matrix rot_matrix about the center of the object.

    If the object has the shape (x, y, z) i.e. is three-dimensional, then the rotation is applied as expected.
    If the shape has the shape (n, x, y, z), i.e. multiple components need to be rotated, the rotation matrix is
    transformed into a 4x4 matrix given by [1, rot_matrix]

    Parameters
    ----------
    config : np.ndarray
        The object to be rotated. Must have shape (x, y, z) or (n, x, y, z)
    rot_matrix : np.ndarray
        The rotation matrix describing the operation to be applied.
    order : int
        The interpolation order of the affine transformation (0 - 5).

    Returns
    -------
    np.ndarray
        The rotated object (same shape as config).
    """
    config_shape = np.asarray(config.shape)

    if config.ndim == 4:
        rot_matrix = block_diag(1, rot_matrix)
    out_center = rot_matrix @ (config_shape - 1) / 2
    in_center = (config_shape - 1) / 2
    offset = in_center - out_center

    if 0 <= order <= 5:
        return affine_transform(config, rot_matrix, offset=offset, order=order)
    else:
        raise Exception("Interpolation order not recognised.")



import numpy as np
import scipy
import magpack.rotations as rtn


def radon(image, thetas=None, order=1):
    """Performs the radon transform on an image.

    Summation / integration occurs along the last dimension. Therefore, an input image with shape (x, y) and (n,) angles
    will result in an output image with shape (n, x). For 3D input of shape (x, y, z) and (n,) the output image will
    have the shape (n, x, y).

    Parameters
    ----------
    image : np.ndarray
        The image to transform.
    thetas : np.ndarray | list (optional)
        Angles to calculate the radon transform.
    order : int (optional)
        Interpolation order.

    Returns
    -------
    np.ndarray
        The Radon transform of the image, the 0th index represents the measured angles.
    """
    thetas = np.asarray(thetas)

    # the first index is the number of angles
    sinogram = np.zeros((thetas.shape[0], *image.shape[:-1]))
    for (ii, theta) in enumerate(thetas):
        # if angles are provided instead of rotation matrices, convert theta array to rotation matrix array
        if not isinstance(theta, np.ndarray):
            theta = rtn.rot(theta) if image.ndim == 2 else rtn.roty(theta)
        sinogram[ii, ...] = rtn.rotate_scalar_field(image, theta, order=order).sum(axis=-1)
    return sinogram


def _get_fourier_filter(size, filter_name):
    """Returns the Fourier filter of specified size (excerpt from scikit-image). """
    n = np.concatenate((np.arange(1, size / 2 + 1, 2, dtype=int), np.arange(size / 2 - 1, 0, -2, dtype=int)))
    f = np.zeros(size)
    f[0] = 0.25
    f[1::2] = -1 / (np.pi * n) ** 2

    fourier_filter = 2 * np.real(np.fft.fft(f))  # ramp filter
    if filter_name == "ramp":
        pass
    elif filter_name == "shepp-logan":
        # Start from the first element to avoid dividing by zero
        omega = np.pi * np.fft.fftfreq(size)[1:]
        fourier_filter[1:] *= np.sin(omega) / omega
    elif filter_name == "cosine":
        freq = np.linspace(0, np.pi, size, endpoint=False)
        cosine_filter = np.fft.fftshift(np.sin(freq))
        fourier_filter *= cosine_filter
    elif filter_name == "hamming":
        fourier_filter *= np.fft.fftshift(np.hamming(size))
    elif filter_name == "hann":
        fourier_filter *= np.fft.fftshift(np.hanning(size))
    elif filter_name is None:
        fourier_filter[:] = 1

    return fourier_filter[:, np.newaxis]


def inv_radon(radon_image, theta=None, filter_name="ramp"):
    """Computes the inverse radon transform of a sinogram / stack of projections.

    Parameters
    ----------
    radon_image : np.ndarray
        The sinogram / stack of projections.
    theta : np.ndarray (optional)
        Angles to calculate the radon transform.
    filter_name : str (optional)
        The name of the filter to use.

    Returns
    -------
    np.ndarray
        The inverse radon transform of the sinogram / stack of projections."""
    if theta is None:
        theta = np.linspace(0, 180, radon_image.shape[0], endpoint=False)
    expanded = False
    parity = 1
    if radon_image.ndim == 2:
        radon_image = radon_image[:, :, np.newaxis]  # odd permutation, changes axes handedness!
        parity = -1
        expanded = True

    if radon_image.ndim != 3:
        raise ValueError("Radon image must be 2D or 3D.")

    angles_count = len(theta)
    if angles_count != radon_image.shape[0]:
        raise ValueError("The given angles do not match the number of projections.")

    filter_types = ('ramp', 'shepp-logan', 'cosine', 'hamming', 'hann', None)
    if filter_name not in filter_types:
        raise ValueError(f"Unknown filter: {filter_name}")

    img_size = radon_image.shape[1]
    projection_size_padded = np.max((64, np.asarray(2 ** np.ceil(np.log2(img_size)), dtype=int)))

    fourier_filter = _get_fourier_filter(projection_size_padded, filter_name)[np.newaxis, ...]
    pad_width = np.asarray([[0, 0], [0, projection_size_padded - img_size], [0, 0]])

    img = np.pad(radon_image, pad_width, mode='constant', constant_values=0)
    projection = np.fft.fft(img, axis=1) * fourier_filter
    radon_filtered = np.real(np.fft.ifft(projection, axis=1))[:, :img_size, ...]
    # Reconstruct image by interpolation
    reconstructed = np.zeros((*radon_image.shape[1:], img_size))
    xpr, zpr = np.mgrid[:img_size, :img_size] - img_size // 2
    x = np.arange(img_size) - img_size // 2

    for plane, angle in zip(radon_filtered, np.deg2rad(theta)):
        t = xpr * np.cos(angle) + parity * zpr * np.sin(angle)
        line_interp = scipy.interpolate.interp1d(x, plane, kind='linear', bounds_error=False, fill_value=0, axis=0)
        reconstructed += line_interp(t).transpose((0, 2, 1))

    final_recons = reconstructed * np.pi / (2 * angles_count)
    final_recons = final_recons[:, 0, :] if expanded else final_recons
    return final_recons




import logging
import numpy as np
import magpack.rotations as rtn
import ThreeDViewer.data
from magtomo.Experiment import Experiment
from magtomo.Reconstruction import Reconstruction
from ThreeDViewer.image import plot_3d
from importlib import resources
from magtomo.radon import radon, inv_radon
from magpack import vectorop, structures, io


def vector_tomo(angles, tilts, kind='CD', it=10, learning=30):
    """ Examples of vector and orientation tomography.

    Parameters
    ----------
    angles : np.ndarray | list of float
        The tomographic rotation angles.
    tilts : np.ndarray | list of float
        The tomographic tilt angles.
    kind : {'CD', 'LD'}
        Vector (CD) or Orientation (LD) tomography options.
    it : int
        The number of gradient descent iterations to evaluate.
    learning : float
        The learning parameter for gradient descent.
    """

    # specify a structure
    struct = _get_structure()
    mask = vectorop.magnitude(struct) > 0

    # get rotation matrices
    rot = rtn.tomo_rot(angles, tilts)
    # get polarizations for each rotation
    pol_a = np.repeat(-45 if kind == 'LD' else 1j, rot.shape[0])
    pol_b = np.repeat(+45 if kind == 'LD' else -1j, rot.shape[0])

    # double rotations to match the polarizations
    pol = np.hstack((pol_a, pol_b))
    rot = np.concatenate((rot, rot), axis=0)
    # initialize the experiment class with the structure, rotation matrices and polarizations
    exp = Experiment(magnetization=struct, rotations=rot, pol=pol)
    # calculate and sinogram that will be used as input
    exp.calculate_sinogram()
    # plot the sinogram
    cmap = 'jet' if kind == 'LD' else 'Spectral'
    exp.plot_sinogram(cmap=cmap)
    # these projections will be used to attempt a reconstruction
    projections = exp.sinogram

    # provide an initial guess for the structure, LD is best with random initial conditions, CD with zero
    initial_guess = np.ones_like(struct) / np.sqrt(3)
    # initialize a reconstruction class and perform reconstruction
    recons = Reconstruction(initial_guess, rotations=rot, projections=projections, pol=pol, iterations=it,
                            mask=mask, learning_parameter=learning)
    recons.reconstruct()
    norm_recons = vectorop.normalize(recons.magnetization)

    sinogram_comparison = np.concatenate([recons.sinogram, exp.sinogram], axis=1)
    a = plot_3d(sinogram_comparison, show=False, cmap=cmap, axes_names=(r"$\theta$", "x", "y"), init_take=0)
    b = plot_3d(norm_recons, axial=True if kind == 'LD' else False, arrow_skip=1, show=False)
    c = plot_3d(struct, axial=True if kind == 'LD' else False, arrow_skip=1)


def scalar_tomo(angles):
    """Example of single axis scalar tomography."""
    # get a scalar field to reconstruct
    struct = _get_structure()[0]

    # calculate projections
    projections = radon(struct, angles)
    plot_3d(projections, cmap='Spectral', init_take=0, axes_names=(r"$\theta$", "x", "y"), title="Projections")

    # reconstruct the object from projections
    recons = inv_radon(projections, angles)

    a = plot_3d(struct, cmap='Spectral', init_take=1, init_slice=47, show=False, title="Initial Structure")
    b = plot_3d(recons, cmap='Spectral', init_take=1, init_slice=47, title="Reconstruction")


def _get_mask(shape):
    """Makes a cylinder mask for the provided (vector field) shape."""
    xx, yy, zz = structures.create_mesh(*shape[-3:])
    r = np.sqrt(xx ** 2 + zz ** 2) < shape[1] / 4  # radius mask based on the x size
    h = shape[2] // 4 > np.abs(yy)  # radius mask based on the y size
    return r * h


def _get_structure():
    """Returns an example magnetization vector field from the ThreeDViewer package."""
    data_path_resource = resources.files(ThreeDViewer.data) / 'cylinder.ovf'
    with data_path_resource as resource:
        data = io.load_ovf(resource).magnetization

    # make y out-of-plane to match the tomography coordinate system
    data = np.array(data[[1, 2, 0]].transpose((0, 2, 3, 1)))
    # mask to leave room for tilting
    return data * _get_mask(data.shape)


def _evenly_spaced(initial, final, steps):
    """Creates an evenly spaced array assuming the ends are cyclic."""
    dx = (final - initial) / steps
    return np.linspace(initial, final - dx, steps)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    initial_angle, final_angle, angle_steps = 0, 360, 18
    rotation_angles = _evenly_spaced(initial_angle, final_angle, angle_steps)

    # example of scalar tomography
    scalar_tomo(rotation_angles)

    # example of vector or orientation tomography
    tilt_angles = [30, -30]
    iteration_number = 10
    learn_param = 30
    dichroism = 'CD'  # circular ('CD') or linear ('LD')
    vector_tomo(rotation_angles, tilt_angles, kind=dichroism, it=iteration_number, learning=learn_param)



import functools
import logging
import numpy as np
import multiprocessing
from typing import Union, List
from ThreeDViewer.image import plot_3d
from magpack.rotations import rotate_scalar_field

Number = Union[int, float, complex]


class Experiment:
    """Defines an experimental setup for performing vector or orientation tomography."""

    def __init__(self, magnetization, rotations, pol=1j, order=1):
        """
        Parameters
        ----------
        magnetization : np.ndarray
            The magnetic configuration to calculate projections from.
        rotations : np.ndarray
            Stack of rotation matrices describing the sample orientations that were measured; shaped (n, 3, 3).
        pol : array_like (optional)
            Single polarization used for all projections, stack of polarizations matching the number of projections or
            stack of polarization pairs. Linear polarization represented using values from 0 to 180 (in degrees).
            Circular left and right polarization represented using complex ±1j.
        order : int
            The interpolation order (0-5) for performing rotations.
        """
        self.magnetization = magnetization
        self.rotations = rotations
        self.pol = pol
        self.order = order
        self._sinogram = None

    @property
    def _is_tensor(self):
        mag_shape = self._magnetization.shape
        if len(mag_shape) == 5 and mag_shape[:2] == (3, 3):
            return True
        else:
            return False

    @property
    def rotations(self):
        """The stack of measured orientation described by rotation matrices."""
        return self._rotations

    @rotations.setter
    def rotations(self, value):
        if value.ndim == 2:
            value = value[np.newaxis, ...]
        if value.shape[-1] != value.shape[-2]:
            raise ValueError("Rotation matrices must be square.")
        if self._magnetization is not None and self._magnetization.shape[0] != value.shape[-1]:
            raise ValueError("Rotation matrix does not match number of field components.")
        self._rotations = value

    @property
    def magnetization(self):
        """The magnetic configuration of the sample."""
        if self._is_tensor:
            self.tensor_to_orientation()
        return self._magnetization

    @magnetization.setter
    def magnetization(self, value: np.ndarray):
        if value.ndim > 5:
            logging.warning(f'WARNING: Invalid magnetization array dimensionality. The array should be either:'
                            f' 4-dimensional with shape (3, nx, ny, nz), i.e. a vector or'
                            f' 5-dimensional with shape (3, 3, nx, ny, nz). Current dimensions: {value.shape}')
        if value.shape[0] != 3:
            logging.warning(f'WARNING: Invalid magnetization components detected. The magnetization array must have'
                            f' exactly 3 components in its first dimension. Current dimensions: {value.shape}')

        self._magnetization = value

    @property
    def pol(self):
        """The polarization of the incident beam."""
        return self._pol

    @pol.setter
    def pol(self, value: Union[Number, List[Number], np.ndarray]):
        if isinstance(value, list):
            value = np.asarray(value)
        self._pol = value

    @property
    def order(self):
        """The interpolation order for performing rotations."""
        return self._order

    @order.setter
    def order(self, value: int):
        if value not in range(5):
            self._order = value
        else:
            self._order = 1

    @property
    def sinogram(self):
        """The calculated sinogram."""
        return self._sinogram

    def orientation_to_tensor(self):
        if not self._is_tensor:
            tensor_mag = np.einsum('iabc,jabc->ijabc', self._magnetization, self._magnetization)
            self.magnetization = tensor_mag

    def tensor_to_orientation(self):
        initial_spatial_dims = self._magnetization.shape[-3:]
        vf_tensor = self._magnetization.reshape(3, 3, -1).transpose(2, 0, 1) # take voxel axis to start
        eig_vals, eig_vecs = np.linalg.eigh(vf_tensor)
        max_indices = np.argmax(eig_vals.real, axis=1)
        vf = np.take_along_axis(eig_vecs, max_indices[:, np.newaxis, np.newaxis], axis=2)[:, :, 0]

        # zero vectors where the max eigenvalue is zero
        mask = eig_vals[np.arange(eig_vals.shape[0]), max_indices].real != 0
        vf[~mask] = 0

        # reshape to match initial spatial dimensions
        self.magnetization = vf.T.reshape(3, *initial_spatial_dims)

    def calculate_sinogram(self):
        """Calculates the sinogram."""
        if isinstance(self._pol, (int, float, complex)):
            self._pol = np.repeat(self._pol, self._rotations.shape[0])

        if np.all(self._pol.imag == 0):
            proj_method = tensor_projection
            if not self._is_tensor:
                self.orientation_to_tensor()
        else:
            proj_method = projection

        partial_projection = functools.partial(proj_method, self._magnetization, order=self._order)
        with multiprocessing.Pool() as p:
            proj = p.starmap(partial_projection, zip(self._rotations, self._pol))
        self._sinogram = np.asarray(proj)

    def plot_sinogram(self, cmap='Spectral'):
        """Plots the calculated sinogram."""
        plot_3d(self._sinogram, init_take=0, cmap=cmap, axes_names=(r"$\theta$", "x", "y"), title="Sinogram")


def get_electric_field(pol):
    """Calculates the electric field vector for given polarization.

    If the polarization is complex, the electric field is assumed to be along the propagation axis (for performing
    vector tomography). If the polarization is real, linear polarization is assumed, with the value of the variable
    being the angle of the polarization in degrees.

    Parameters
    ----------
    pol : float or complex

    Returns
    -------
    array_like
    """
    if isinstance(pol, complex) and pol.imag != 0:
        return np.array([0, 0, pol.imag])  # Assume imaginary pol is along z
    else:
        pol = pol.real if isinstance(pol, complex) else pol
        return np.array([np.cos(np.deg2rad(pol)), np.sin(np.deg2rad(pol)), 0])

def tensor_projection(tensor_field, rot_matrix, pol=0, order=1):
    """Calculation of a single projection using tensor notation

    Parameters
    ----------
    tensor_field : np.ndarray
        The magnetization tensor field.
    rot_matrix : np.ndarray
        A single rotation matrix.
    pol : np.ndarray (optional)
        Single polarization or a pair of polarizations (if the difference of projections is desired). Linear
        polarization represented using values from 0 to 180 (in degrees) and circular polarizations represented with
        +1j and -1j (complex unit values)
    order : int (optional)
        The interpolation order (0-5) for performing rotations.

    Returns
    -------
    np.ndarray
        The projection of the field at the orientation described by `rot_matrix` and polarization `pol`.
    """
    field_copy = np.copy(tensor_field)

    electric_field = (np.cos(np.deg2rad(pol)), np.sin(np.deg2rad(pol)), 0)
    proj = np.einsum('i,ij,jlabc,kl,k->abc', electric_field, rot_matrix, field_copy, rot_matrix,
                     electric_field)

    proj = rotate_scalar_field(proj, rot_matrix, order=order)
    return np.sum(proj, axis=2)


def projection(field, rot_matrix, pol=0, order=1):
    """Calculation of a single projection.

    Parameters
    ----------
    field : np.ndarray
        The magnetization field.
    rot_matrix : np.ndarray
        A single rotation matrix.
    pol : np.ndarray (optional)
        Single polarization or a pair of polarizations (if the difference of projections is desired). Linear
        polarization represented using values from 0 to 180 (in degrees) and circular polarizations represented with
        +1j and -1j (complex unit values)
    order : int (optional)
        The interpolation order (0-5) for performing rotations.

    Returns
    -------
    np.ndarray
        The projection of the field at the orientation described by `rot_matrix` and polarization `pol`.
    """
    field_copy = np.copy(field)

    if isinstance(pol, np.ndarray) and len(pol) == 2:
        return projection(field, rot_matrix, pol[0], order) - projection(field, rot_matrix, pol[1], order)
    elif isinstance(pol, np.ndarray):
        raise ValueError("Polarisations for dichroic input must be in pairs.")

    electric_field = get_electric_field(pol)
    if electric_field[2] != 0:
        proj = np.einsum('j,jabc -> abc', rot_matrix[2], field_copy) * electric_field[2]
    else:
        proj = np.einsum('ij,jabc,i->abc', rot_matrix, field_copy, electric_field) ** 2

    proj = rotate_scalar_field(proj, rot_matrix, order=order)
    return np.sum(proj, axis=2)
